"""
    TCP 服务端

    菜鸟教程
    https://www.runoob.com/python/python-socket.html

    问题 1：粘包
        描述：即发送消息，接收方不知道单次发送大小，导致接收方一次接收到发送方的多次消息，导致粘包

        解决：使用 struct 模块，每次发送信息前，都发送一下当前数据的长度，再去接收
"""
import socket
import threading
import time
from pathlib import Path
from typing import List, Type
from loguru import logger
from gsocket.sender import SocketSender
from gsocket.receiver import SocketReceiver


class ServerHousekeeper(threading.Thread):
    def __init__(
            self,
            client: socket.socket,
            clients: List[socket.socket],
            size: int,
            timeout: int,
            receiver: SocketReceiver,
            encoding: str = None
    ):
        """
        服务于每个客户端连接的管家

        :param client: 当前连接客户端
        :param clients: 所有的客户端
        :param size: 接收的大小
        :param timeout: 超时时间
        :param receiver: 信息接收器
        :param encoding: 编码格式
        """
        super().__init__()
        self.client = client
        self.clients = clients
        self.size = size
        self.timeout = timeout
        self.receiver = receiver
        self.encoding = encoding or 'utf-8'

        self.address = self.client.getpeername()  # 客户端的 ip、port
        self.sender = self.receiver.sender

    def run(self) -> None:
        """
        客户端循环监听

        :return:
        """
        # 启动函数
        self.on_open()

        try:
            # 监听消息
            self.receiver.start()

            # 等待执行结束
            self.waiting_done()
        finally:
            # 关闭函数
            self.on_close()

    def send(self, msg: str):
        """
        发送消息

        :param msg:
        :return:
        """
        self.sender.send(msg=msg)

    def sendall(self, msg: str):
        """
        发送消息

        :param msg:
        :return:
        """
        self.sender.sendall(msg=msg)

    def sendfile(self, filepath: str):
        """
        发送文件

        :param filepath:
        :return:
        """
        self.sender.sendfile(filepath=filepath)

    def waiting_done(self):
        """
        等待接收、发送线程执行完毕

        :return:
        """
        while True:
            if not self.receiver.is_alive():
                break

            time.sleep(0.1)

    def on_open(self):
        """
        客户端建立连接时

        :return:
        """
        self.clients.append(self.client)
        if self.client.gettimeout() is None:
            self.client.settimeout(self.timeout)
        logger.debug(f'建立新连接：{self.address}')

    def on_close(self):
        """
        客户端关闭时

        :return:
        """
        self.client.close()
        self.clients.remove(self.client)
        logger.debug(f"已断开连接：{self.address}")


class SocketServer:
    def __init__(
            self,
            host: str = None,
            port: int = None,
            listen: int = None,
            size: int = None,
            timeout: int = None,
            receiver: Type[SocketReceiver] = None,
            file_recv_path: str = None,
            encoding: str = None
    ):
        """

        :param host: host
        :param port: port
        :param listen: 一个服务同时处理几个
        :param size: 接受文件切片大小
        :param timeout: 超时时间
        :param receiver: 信息接收器
        :param file_recv_path: 文件接收路径
        :param encoding: 编码格式
        """
        self.host = host or '127.0.0.1'
        self.port = port or 8866
        self.listen = listen or 5
        self.size = size or 1024
        self.timeout = timeout or 30
        self.receiver = receiver or SocketReceiver
        self.file_recv_path = file_recv_path or Path(__file__).parent
        self.encoding = encoding or 'utf-8'

        self.server = None  # 服务端
        self.clients = []  # 存储所有连接的客户端

    def start_server(self):
        """
        开启 TCP 服务

        :return:
        """
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind((self.host, self.port))
        self.server.listen(self.listen)

        logger.debug(f"服务端已启动，当前监听：{self.server.getsockname()}")

        while True:
            client, addr = self.server.accept()
            receiver = self.receiver(
                client=client,
                size=self.size,
                sender=SocketSender(client=client),
                file_recv_path=self.file_recv_path,
                encoding=self.encoding
            )
            ServerHousekeeper(
                client=client,
                clients=self.clients,
                size=self.size,
                timeout=self.timeout,
                receiver=receiver,
                encoding=self.encoding
            ).start()

    def __del__(self):
        self.server.close()


if __name__ == '__main__':
    SocketServer().start_server()



"""
    TCP 客户端

    示例1：
        tcp = SocketClient()
        tcp.send(msg="你好啊，今天是10.05号")
        tcp.sendall(msg="你好啊，今天是10.05号")
        tcp.close()

    示例2：
        with SocketClient() as tcp:
            tcp.send(msg="你好啊，今天是10.05号")
            tcp.sendall(msg="你好啊，今天是10.05号")
            tcp.sendfile(r'D:\Download\wallpaper\wallhaven-3lo8q3_1920x1080.png')
"""
import socket
from typing import Type
from pathlib import Path
from gsocket import singal
from gsocket.sender import SocketSender
from gsocket.receiver import SocketReceiver


class SocketClient:
    def __init__(
            self,
            host: str = None,
            port: int = None,
            size: int = None,
            timeout: int = None,
            receiver: Type[SocketReceiver] = None,
            file_recv_path: str = None,
            encoding: str = None,
    ):
        """

        :param host: host
        :param port: port
        :param size: 单次接收大小
        :param timeout: 超时时间
        :param receiver: 信息接收器
        :param file_recv_path: 文件接收路径
        :param encoding: 编码格式
        """
        self.host = host or '127.0.0.1'
        self.port = port or 8866
        self.size = size or 1024
        self.timeout = timeout or 30
        receiver = receiver or SocketReceiver
        self.file_recv_path = file_recv_path or Path(__file__).parent
        self.encoding = encoding

        # 创建连接
        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.client.connect((self.host, self.port))
        self.client.settimeout(self.timeout)
        self.closed = False
        self.sender = SocketSender(client=self.client)
        self.receiver = receiver(
            client=self.client,
            size=self.size,
            sender=self.sender,
            file_recv_path=self.file_recv_path,
            encoding=self.encoding
        )
        self.receiver.start()

    def connect(self):
        """
        创建连接

        :return:
        """

        return self

    def send(self, msg: str):
        """
        发送消息

        :param msg:
        :return:
        """
        self.sender.send(msg=msg)

    def sendall(self, msg: str):
        """
        发送消息

        :param msg:
        :return:
        """
        self.sender.sendall(msg=msg)

    def sendfile(self, filepath: str):
        """
        发送文件

        :param filepath:
        :return:
        """
        self.sender.sendfile(filepath=filepath)

    def _send_close_signal(self):
        """
        发送结束信号，让服务器释放连接

        :return:
        """
        self.send(singal.CLOSE)

        while True:
            if not self.receiver.is_alive():
                break

    def close(self):
        if not self.closed:
            self._send_close_signal()
            self.client.close()
            self.closed = True

    def __enter__(self):
        self.connect()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
    接收器

    注意：这里使用了 struct 先发送消息头，才会去发送消息，所以其它的程序发送的，没有头的话可能接收不到
"""
import socket
import struct
import threading
from pathlib import Path
from loguru import logger
from gsocket import singal
from typing import Generator
from gsocket.sender import SocketSender


class SocketReceiver(threading.Thread):
    def __init__(self, client: socket.socket, sender: SocketSender, size: int = None, file_recv_path: str = None,
                 encoding: str = None):
        """

        :param client: socket 连接
        :param size: 一次接收的最大大小
        :param sender: 发送器
        :param file_recv_path: 文件接收路径
        :param encoding: 编码格式
        """
        super().__init__(daemon=True)
        self.client = client
        self.size = size or 1024
        self.sender = sender
        self.file_recv_path = Path(file_recv_path)
        self.encoding = encoding or 'utf-8'

    def run(self) -> None:
        """
        监控数据接收

        :return:
        """
        try:
            while True:
                header = self.client.recv(4)  # 获取报头

                print(header)
                print(header.decode())
                if header == b'':
                    break

                msg_cache = self.recv_msg(header)

                # 判断是文件还是信息，分别发送到不同的接收函数
                if msg_cache == singal.File:
                    file_receiver = self.recv_file()
                    filename = next(file_receiver)
                    self.on_file(filename, file_receiver)
                elif msg_cache == singal.CLOSE:
                    break
                else:
                    self.on_message(msg_cache)
        except ConnectionError:
            logger.warning(f"远程主机断开：{self.client.getpeername()}")

    def recv_msg(self, header: bytes) -> bytes:
        """
        接收信息

        :param header: 消息头
        :return:
        """
        # 获取待接收数据长度
        length = struct.unpack('i', header)[0]

        # 计算每次接收的长度
        if length <= self.size:
            index = length
        else:
            index = self.size

        # 开始接收数据
        msg_cache = b''  # 已接收的数据
        received_length = 0  # 已接收长度
        while True:
            msg = self.client.recv(index)
            msg_cache += msg

            # 累加已接收长度
            received_length += len(msg)
            if received_length == length:
                break

            # 判断下次接收长度（针对这里 self.size 接收的比较小的情况）
            if length - received_length < self.size:
                index = length - received_length

        return msg_cache

    def recv_file(self):
        """
        接收信息，返回的是文件内容的迭代器

        :return:
        """
        file_header = struct.calcsize('128sl')
        file_buffer = self.client.recv(file_header)

        if file_buffer:
            filename, filesize = struct.unpack('128sl', file_buffer)
            filename = filename.strip(b'\x00').decode()

            yield filename

            filesize_count = 0
            while True:
                header = self.client.recv(4)
                data = self.recv_msg(header=header)
                filesize_count += len(data)

                if data:
                    yield data

                if filesize_count >= filesize:
                    break

    def send(self, msg: str):
        """
        发送消息

        :param msg:
        :return:
        """
        self.sender.send(msg=msg)

    def sendall(self, msg: str):
        """
        发送消息

        :param msg:
        :return:
        """
        self.sender.sendall(msg=msg)

    def sendfile(self, filepath: str):
        """
        发送文件

        :param filepath:
        :return:
        """
        self.sender.sendfile(filepath=filepath)

    def on_open(self):
        """
        客户端建立连接时

        :return:
        """
        pass

    def on_message(self, msg: bytes):
        """
        接收到消息时

        :param msg:
        :return:
        """
        logger.info(msg.decode(self.encoding))

    def on_file(self, filename: str, file_iterator: Generator):
        """
        接收到文件时

        :param filename: 文件名
        :param file_iterator: 文件内容生成器
        :return:
        """
        filepath = self.file_recv_path.joinpath(filename)
        with open(filepath, 'wb') as f:
            for file_buffer in file_iterator:
                f.write(file_buffer)

        logger.debug(f"文件已保存到本地：{filepath}")

    def on_close(self):
        """
        客户端关闭时

        :return:
        """
        pass



"""
    发送器

    注意：这里使用了 struct 先发送消息头，才会去发送消息
"""
import os
import socket
import struct
from typing import Union
from gsocket import singal


class SocketSender:
    def __init__(self, client: socket.socket, size: int = None, encoding: str = None):
        """

        :param client: 连接
        :param size: 单次最大发送大小
        :param encoding: 编码格式
        """
        self.client = client
        self.size = size or 1024
        self.encoding = encoding or 'utf-8'

    def send(self, msg: Union[str, bytes]):
        """
        发送 TCP 消息
        并非是一次性发送完，所以要接收剩余长度

        :param msg:
        :return:
        """
        if isinstance(msg, str):
            msg = msg.encode(self.encoding)

        index = 0
        length = len(msg)

        while True:
            msg_send = msg[index:]
            self.send_msg_header(msg=msg_send)
            index = self.client.send(msg_send)  # 返回发送的长度

            if index < length:
                index += 1
            else:
                break

    def sendall(self, msg: Union[str, bytes]):
        """
        完整的发送 TCP 消息（推荐）

        :param msg:
        :return:
        """
        if isinstance(msg, str):
            msg = msg.encode(self.encoding)

        self.send_msg_header(msg=msg)
        self.client.sendall(msg)

    def sendfile(self, filepath: str):
        """
        发送文件

        :param filepath:
        :return:
        """
        self.send_file_header(filepath=filepath)

        with open(filepath, 'rb') as f:
            while True:
                data = f.read(self.size)
                if not data:
                    break

                self.sendall(msg=data)

    def send_msg_header(self, msg: bytes):
        """
        创建消息头

        :param msg:
        :return:
        """
        # 先发送标志
        header = struct.pack('i', len(msg))
        self.client.sendall(header)

    def send_file_header(self, filepath: str):
        """
        发送文件头

        :param filepath: 文件路径
        :return:
        """
        # 先告诉其将发送文件
        self.sendall(msg=singal.File)

        # 定义文件头信息（包含文件名和文件大小）
        file_header = struct.pack('128sl', os.path.basename(filepath).encode(self.encoding), os.stat(filepath).st_size)

        # 发送文件头
        self.client.sendall(file_header)
